#include "llvm/Pass.h"
#include "llvm/IR/Function.h"
#include "llvm/IR/Module.h"
#include "llvm/IR/BasicBlock.h"
#include "llvm/IR/Instructions.h"
#include "llvm/IR/LegacyPassManager.h"
#include "llvm/Transforms/IPO/PassManagerBuilder.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Analysis/CFG.h"
#include "llvm/Analysis/CallGraph.h"
#include "LogSystem.h"
#include <vector>
#include <map>
#include <set>
#include <queue>
#include <random>
#include <algorithm>
#include <string>

using namespace llvm;

namespace {
    struct CallNode {
        std::string name;
        bool isTarget;
        std::set<std::string> callers;
        std::set<std::string> callees;
        unsigned depth;
        std::set<BasicBlock*> criticalPoints;
        unsigned num; // 融合点个数/分片个数
        
        CallNode() : name(""), isTarget(false), depth(0), num(0) {}

        CallNode(std::string n, bool target = false) 
            : name(n), isTarget(target), depth(0), num(0) {}
    };

    struct CodeFusionPass : public ModulePass {
    public:
        static char ID;
        CodeFusionPass() : ModulePass(ID) {}

        bool runOnModule(Module &M) override {
            auto& logger = logging::LogSystem::getInstance();
            logger.setGlobalLevel(logging::LogLevel::DEBUG);
            // 配置Log输出权限
            logger.enableFunction("runOnModule");
            // logger.enableFunction("dumpControlFlowGraph");
            logger.enableFunction("buildCallGraph");
            // logger.enableFunction("analyzeFunctionCriticalPoints");
            // logger.enableFunction("calculateCallDepths");
            logger.enableFunction("matchFunctionsForFusion");
            logger.enableFunction("createFunctionSlices");
            logger.enableFunction("countFusionPoints");
            logger.enableFunction("generateFusionMatchGraph");
            
            LOG_INFO("runOnModule", "Starting analysis for module: {0}", M.getName().str());

            for (Function &F : M) {
                if(!F.isDeclaration()){
                    if (isTargetCode(F)) {
                        targetFunctions.insert(F.getName().str());
                    }
                    else{
                        bunkerFunctions.insert(F.getName().str());
                    }
                }
            }

            buildCallGraph(M);
            analyzeFunctionCriticalPoints(M);
            calculateCallDepths(M);
            generateProjectCallGraph();
            for (Function &F : M) {
                if(!F.isDeclaration()){
                    dumpControlFlowGraph(F);
                }
            }

            LOG_INFO("runOnModule", "Creating slices for target functions");
            for (const auto& targetFuncName : targetFunctions) {
                LOG_INFO("runOnModule", "targetFuncName: {0}",targetFuncName);
                if (Function* F = M.getFunction(targetFuncName)) {
                    createFunctionSlices(*F);
                } else {
                    LOG_ERROR("runOnModule", "Could not find function {0} in module", targetFuncName);
                }
            }

            LOG_INFO("runOnModule", "Creating fusion points for bunker functions");
            for (const auto& bunkerFuncName : bunkerFunctions) {
                LOG_INFO("runOnModule", "bunkerFuncName: {0}",bunkerFuncName);
                if (Function* F = M.getFunction(bunkerFuncName)) {
                    countFusionPoints(*F);
                } else {
                    LOG_ERROR("runOnModule", "Could not find function {0} in module", bunkerFuncName);
                }
            }

            if(matchFunctionsForFusion()){
                generateFusionMatchGraph();
            }

            return false;
        }

    private:
        std::map<std::string, CallNode> callGraph;
        std::map<std::string, unsigned> maxCallDepths;
        std::set<std::string> targetFunctions;
        std::set<std::string> bunkerFunctions;
        std::map<std::string, std::string> fusionPairs;

        struct Slice {
            std::vector<BasicBlock*> blocks;
        };

        size_t getRandomIndex(size_t max) {
            static unsigned seed = 0;
            seed = (seed * 1103515245 + 12345) & 0x7fffffff;
            return seed % max;
        }

        bool matchFunctionsForFusion() {
            LOG_INFO("matchFunctionsForFusion", "Starting function matching process");
            
            if (bunkerFunctions.size() < targetFunctions.size()) {
                LOG_WARNING("matchFunctionsForFusion", 
                    "Insufficient bunker functions ({0}) for target functions ({1})", 
                    bunkerFunctions.size(), 
                    targetFunctions.size());
                return false;
            }

            std::vector<std::string> sortedTargets(targetFunctions.begin(), targetFunctions.end());
            std::sort(sortedTargets.begin(), sortedTargets.end(),
                [this](const std::string& a, const std::string& b) {
                    return callGraph[a].num > callGraph[b].num;  // 降序排列
                });

            std::vector<std::string> availableBunkers(bunkerFunctions.begin(), bunkerFunctions.end());

            for (const auto& targetFunc : sortedTargets) {
                const auto& targetNode = callGraph[targetFunc];
                size_t requiredPoints = targetNode.num;

                bool matched = false;
                while (!availableBunkers.empty() && !matched) {
                    size_t idx = getRandomIndex(availableBunkers.size());
                    const auto& bunkerFunc = availableBunkers[idx];
                    const auto& bunkerNode = callGraph[bunkerFunc];

                    if (bunkerNode.num >= requiredPoints) {
                        fusionPairs[targetFunc] = bunkerFunc;
                        availableBunkers.erase(availableBunkers.begin() + idx);
                        matched = true;
                        LOG_INFO("matchFunctionsForFusion", 
                            "Matched target {0} ({1} slices) with bunker {2} ({3} fusion points)",
                            targetFunc, 
                            targetNode.num,
                            bunkerFunc,
                            bunkerNode.num);
                    } else {
                        availableBunkers.erase(availableBunkers.begin() + idx);
                    }
                }

                if (!matched) {
                    LOG_ERROR("matchFunctionsForFusion", 
                        "Could not find suitable bunker function for target {0}", 
                        targetFunc);
                    return false;
                }
                // LOG_INFO("matchFunctionsForFusion", "Remaining bunker functions after matching {0}:", targetFunc);
                // for (const auto& remainingBunker : availableBunkers) {
                //     const auto& bunkerNode = callGraph[remainingBunker];
                //     LOG_INFO("matchFunctionsForFusion", 
                //         "    {0} (Critical Points: {1})", 
                //         remainingBunker, 
                //         bunkerNode.criticalPoints.size());
                // }
            }
            return true;
        }

        void generateFusionMatchGraph() {
            LOG_INFO("generateFusionMatchGraph", "Generating fusion match visualization");
            
            errs() << "```mermaid\n";
            errs() << "graph LR\n";
            errs() << "    %% Fusion Match Graph\n";
            errs() << "    %% Left side: Target Project, Right side: Bunker Project\n\n";
            
            errs() << "    subgraph Target[\"Target Project\"]\n";
            errs() << "        direction TB\n";
            for (const auto& targetFunc : targetFunctions) {
                std::string nodeId = sanitizeNodeId(targetFunc);
                std::string criticalPoints = std::to_string(callGraph[targetFunc].criticalPoints.size());
                errs() << "        " << nodeId << "[\"" << targetFunc 
                    << "\\nCritical Points: " << criticalPoints
                    << "\"]:::target\n";
            }
            errs() << "    end\n\n";
            
            errs() << "    subgraph Bunker[\"Bunker Project\"]\n";
            errs() << "        direction TB\n";
            for (const auto& bunkerFunc : bunkerFunctions) {
                std::string nodeId = sanitizeNodeId(bunkerFunc);
                std::string criticalPoints = std::to_string(callGraph[bunkerFunc].criticalPoints.size());
                errs() << "        " << nodeId << "[\"" << bunkerFunc 
                    << "\\nCritical Points: " << criticalPoints
                    << "\"]:::bunker\n";
            }
            errs() << "    end\n\n";
            
            errs() << "    %% Fusion Matches\n";
            for (const auto& match : fusionPairs) {
                std::string targetId = sanitizeNodeId(match.first);
                std::string bunkerId = sanitizeNodeId(match.second);
                errs() << "    " << targetId << " ==>|fusion| " << bunkerId << "\n";
            }
            
            errs() << "    %% Styles\n";
            errs() << "    classDef target fill:#f96,stroke:#333,stroke-width:2px\n";
            errs() << "    classDef bunker fill:#9cf,stroke:#333,stroke-width:2px\n";
            errs() << "    style Target fill:#fff,stroke:#f96,stroke-width:2px\n";
            errs() << "    style Bunker fill:#fff,stroke:#9cf,stroke-width:2px\n";
            
            errs() << "```\n\n";
            
            LOG_INFO("generateFusionMatchGraph", 
                "Generated visualization with {0} target functions and {1} bunker functions", 
                targetFunctions.size(), 
                bunkerFunctions.size());
            LOG_INFO("generateFusionMatchGraph", 
                "Total fusion matches: {0}", 
                fusionPairs.size());
        }

        void countFusionPoints(Function& F) {
            std::string fname = F.getName().str();
            
            if (targetFunctions.find(fname) != targetFunctions.end()) {
                return;
            }

            auto& criticalPoints = callGraph[fname].criticalPoints;
            if (criticalPoints.empty()) {
                callGraph[fname].num = 0;
                return;
            }
            callGraph[fname].num = criticalPoints.size();

            LOG_INFO("countFusionPoints", 
                "Counted {0} fusion points for bunker function {1}",
                criticalPoints.size(),
                fname);
        }

        void createFunctionSlices(Function& F) {
            std::string fname = F.getName().str();
            
            if (!callGraph[fname].isTarget) {
                return;
            }

            auto& criticalPoints = callGraph[fname].criticalPoints;
            if (criticalPoints.empty()) {
                return;
            }

            BasicBlock* entryBlock = &F.getEntryBlock();
            criticalPoints.insert(entryBlock);

            std::vector<BasicBlock*> orderedPoints(criticalPoints.begin(), criticalPoints.end());
            std::sort(orderedPoints.begin(), orderedPoints.end(),
                [](BasicBlock* a, BasicBlock* b) {
                    return a->getName() < b->getName();
                });

            std::set<BasicBlock*> processedBlocks;
            std::vector<Slice> slices;

            for (size_t i = 0; i < orderedPoints.size(); i++) {
                BasicBlock* currentCritical = orderedPoints[i];
                
                if (processedBlocks.count(currentCritical)) {
                    continue;
                }

                size_t numSuccessors = std::distance(succ_begin(currentCritical), succ_end(currentCritical));
                size_t numPredecessors = std::distance(pred_begin(currentCritical), pred_end(currentCritical));

                Slice newSlice;
                if (numSuccessors <= 1 && numPredecessors <= 1) {
                    newSlice.blocks.push_back(currentCritical);
                    processedBlocks.insert(currentCritical);
                } else {
                    BasicBlock* nextCritical = nullptr;
                    for (size_t j = i + 1; j < orderedPoints.size(); j++) {
                        if (!processedBlocks.count(orderedPoints[j])) {
                            nextCritical = orderedPoints[j];
                            break;
                        }
                    }

                    if (!nextCritical) {
                        newSlice.blocks.push_back(currentCritical);
                        processedBlocks.insert(currentCritical);
                    } else {
                        std::queue<BasicBlock*> queue;
                        queue.push(currentCritical);
                        while (!queue.empty()) {
                            BasicBlock* current = queue.front();
                            queue.pop();
                            if (processedBlocks.count(current)) {
                                continue;
                            }
                            processedBlocks.insert(current);
                            newSlice.blocks.push_back(current);
                            if (current != nextCritical) {
                                for (BasicBlock* succ : successors(current)) {
                                    if (!processedBlocks.count(succ) && 
                                        (succ == nextCritical || !criticalPoints.count(succ))) {
                                        queue.push(succ);
                                    }
                                }
                            }
                        }
                    }
                }
                if (!newSlice.blocks.empty()) {
                    slices.push_back(newSlice);
                }
            }

            callGraph[fname].num = slices.size();

            LOG_INFO("createFunctionSlices", 
                "Created {0} slices for target function {1}",
                slices.size(),
                fname);

            for (size_t i = 0; i < slices.size(); i++) {
                std::string blockNames;
                for (BasicBlock* block : slices[i].blocks) {
                    if (!blockNames.empty()) {
                        blockNames += ", ";
                    }
                    blockNames += getNodeId(block);
                }
                LOG_DEBUG("createFunctionSlices", 
                    "Slice {0} of function {1} contains blocks: {2}",
                    i,
                    fname,
                    blockNames);
            }
        }

        bool isTargetCode(Function &F) {
            if (MDNode *MD = F.getMetadata("project_source")) {
                if (MDString *ProjectStr = dyn_cast<MDString>(MD->getOperand(0))) {
                    std::string projectName = ProjectStr->getString().str();
                    return (projectName == "Target");
                }
            }
            return false;
        }

        void dumpControlFlowGraph(Function &F) {
            LOG_INFO("dumpControlFlowGraph", "Generating control flow graph for function: {0}", 
                F.getName().str());
            
            if (F.empty()) {
                LOG_WARNING("dumpControlFlowGraph", "Function is empty!");
                return;
            }

            LOG_INFO("dumpControlFlowGraph", "Starting Mermaid graph generation");
            errs() << "```mermaid: " << F.getName().str() << "\n";
            errs() << "graph TD\n";
            
            for (BasicBlock &BB : F) {
                std::string nodeId = getNodeId(&BB);
                if (!nodeId.empty() && nodeId[0] == '%') {
                    nodeId = nodeId.substr(1);
                }
                std::replace(nodeId.begin(), nodeId.end(), '.', '_');
                std::replace(nodeId.begin(), nodeId.end(), ' ', '_');
                std::replace(nodeId.begin(), nodeId.end(), '%', '_');
                std::replace(nodeId.begin(), nodeId.end(), '-', '_');

                LOG_TRACE("dumpControlFlowGraph", "Processing block with ID: {0}", nodeId);

                std::string blockContent;
                raw_string_ostream contentOS(blockContent);
                
                contentOS << "Block " << nodeId << ":\\n";
                
                for (Instruction &I : BB) {
                    std::string instStr;
                    raw_string_ostream instOS(instStr);
                    I.print(instOS);
                    instOS.flush();
                    
                    std::replace(instStr.begin(), instStr.end(), '"', '\'');
                    std::replace(instStr.begin(), instStr.end(), '\n', ' ');
                    
                    if (instStr.length() > 50) {
                        LOG_TRACE("dumpControlFlowGraph", "Truncating long instruction in block {0}", nodeId);
                        instStr = instStr.substr(0, 47) + "...";
                    }
                    
                    contentOS << instStr << "\\n";
                }
                contentOS.flush();

                std::string fname = F.getName().str();

                if (callGraph[fname].criticalPoints.count(&BB)) {
                    LOG_TRACE("dumpControlFlowGraph", "Marking block {0} as critical", nodeId);
                    errs() << "    " << nodeId << "[\"" << blockContent << "\"]:::critical\n";
                } else {
                    errs() << "    " << nodeId << "[\"" << blockContent << "\"]\n";
                }

                for (BasicBlock *Succ : successors(&BB)) {
                    std::string succId;
                    raw_string_ostream succOS(succId);
                    Succ->printAsOperand(succOS, false);
                    succOS.flush();
                    
                    if (!succId.empty() && succId[0] == '%') {
                        succId = succId.substr(1);
                    }
                    
                    std::replace(succId.begin(), succId.end(), '.', '_');
                    std::replace(succId.begin(), succId.end(), ' ', '_');
                    std::replace(succId.begin(), succId.end(), '%', '_');
                    std::replace(succId.begin(), succId.end(), '-', '_');

                    LOG_TRACE("dumpControlFlowGraph", "Processing edge from {0} to {1}", nodeId, succId);

                    if (BranchInst *BI = dyn_cast<BranchInst>(BB.getTerminator())) {
                        if (BI->isConditional()) {
                            std::string condStr;
                            raw_string_ostream condOS(condStr);
                            BI->getCondition()->print(condOS);
                            condOS.flush();
                            
                            bool isTrue = BI->getSuccessor(0) == Succ;
                            LOG_TRACE("dumpControlFlowGraph", "Adding conditional branch edge: {0}", 
                                isTrue ? "true" : "false");
                            errs() << "    " << nodeId << " -->|" 
                                << (isTrue ? "true" : "false") << "| " 
                                << succId << "\n";
                        } else {
                            errs() << "    " << nodeId << " --> " << succId << "\n";
                        }
                    } else if (SwitchInst *SI = dyn_cast<SwitchInst>(BB.getTerminator())) {
                        if (Succ == SI->getDefaultDest()) {
                            LOG_TRACE("dumpControlFlowGraph", "Adding switch default edge");
                            errs() << "    " << nodeId << " -->|default| " << succId << "\n";
                        } else {
                            for (auto Case : SI->cases()) {
                                if (Case.getCaseSuccessor() == Succ) {
                                    std::string caseStr;
                                    raw_string_ostream caseOS(caseStr);
                                    Case.getCaseValue()->print(caseOS);
                                    LOG_TRACE("dumpControlFlowGraph", "Adding switch case edge: {0}", 
                                        caseStr);
                                    errs() << "    " << nodeId << " -->|case " 
                                        << caseStr << "| " << succId << "\n";
                                }
                            }
                        }
                    } else {
                        errs() << "    " << nodeId << " --> " << succId << "\n";
                    }
                }
            }
            errs() << "    classDef critical fill:#f96,stroke:#333,stroke-width:4px\n";
            errs() << "```\n";
            LOG_INFO("dumpControlFlowGraph", "Completed graph generation");
        }

        std::string getNodeId(BasicBlock *BB) {
            std::string nodeId;
            raw_string_ostream idOS(nodeId);
            BB->printAsOperand(idOS, false);
            idOS.flush();
            return nodeId;
        }

        bool checkCriticalPoint(BasicBlock *BB) {
            if (!BB) return false;
            Function *F = BB->getParent();
            if (!F) return false;
            BasicBlock *Entry = &F->getEntryBlock();
            BasicBlock *Exit = nullptr;
            for (BasicBlock &B : *F) {
                if (isa<ReturnInst>(B.getTerminator()) || 
                    isa<UnreachableInst>(B.getTerminator())) {
                    Exit = &B;
                    break;
                }
            }
            if (!Exit) return false;
            std::set<BasicBlock*> visitedWithoutCurrent;
            std::function<bool(BasicBlock*, BasicBlock*)> canReachExitWithout;
            canReachExitWithout = [&canReachExitWithout, &visitedWithoutCurrent, BB]
                (BasicBlock *Start, BasicBlock *Target) -> bool {
                    if (Start == BB) return false;
                    if (Start == Target) return true;
                    
                    if (visitedWithoutCurrent.count(Start)) return false;
                    visitedWithoutCurrent.insert(Start);

                    for (BasicBlock *Succ : successors(Start)) {
                        if (canReachExitWithout(Succ, Target)) {
                            return true;
                        }
                    }
                    return false;
                };
            visitedWithoutCurrent.clear();
            return !canReachExitWithout(Entry, Exit);
        }

        void analyzeFunctionCriticalPoints(Module &M) {
            for (Function &F : M) {
                if (!F.isDeclaration()) {
                    std::string fname = F.getName().str();
                    
                    for (BasicBlock &BB : F) {
                        if (checkCriticalPoint(&BB)) {
                            callGraph[fname].criticalPoints.insert(&BB);
                        }
                    }
                    LOG_INFO("analyzeFunctionCriticalPoints", 
                        "Function {0} has {1} critical points", 
                        fname, 
                        callGraph[fname].criticalPoints.size());
                }
            }
        }

        void buildCallGraph(Module &M) {
            LOG_INFO("buildCallGraph", "Building complete call graph");
            for (Function &F : M) {
                if (!F.isDeclaration()) {
                    std::string fname = F.getName().str();
                    bool isTarget = targetFunctions.find(fname) != targetFunctions.end();
                    callGraph[fname] = CallNode(fname, isTarget);
                }
            }
            for (Function &F : M) {
                if (!F.isDeclaration()) {
                    std::string callerName = F.getName().str();
                    for (BasicBlock &BB : F) {
                        for (Instruction &I : BB) {
                            if (CallInst *CI = dyn_cast<CallInst>(&I)) {
                                Function *CalledF = CI->getCalledFunction();
                                if (CalledF && !CalledF->isDeclaration()) {
                                    std::string calleeName = CalledF->getName().str();
                                    callGraph[callerName].callees.insert(calleeName);
                                    callGraph[calleeName].callers.insert(callerName);
                                }
                            }
                        }
                    }
                }
            }
        }

        void calculateCallDepths(Module &M) {
            LOG_INFO("calculateCallDepths", "Calculating call depths for all functions in module");

            maxCallDepths.clear();
            
            for (Function &F : M) {
                if (F.isDeclaration()) {
                    continue;
                }
                std::string startFunc = F.getName().str();
                std::map<std::string, unsigned> depths;
                std::queue<std::pair<std::string, unsigned>> queue;
                std::set<std::string> visited;

                queue.push(std::make_pair(startFunc, 0));
                visited.insert(startFunc);

                while (!queue.empty()) {
                    std::string currentFunc = queue.front().first;
                    unsigned depth = queue.front().second;
                    queue.pop();
                    depths[currentFunc] = depth;
                    for (const std::string &caller : callGraph[currentFunc].callers) {
                        if (visited.find(caller) == visited.end()) {
                            queue.push(std::make_pair(caller, depth + 1));
                            visited.insert(caller);
                        }
                    }
                    for (const std::string &callee : callGraph[currentFunc].callees) {
                        if (visited.find(callee) == visited.end()) {
                            queue.push(std::make_pair(callee, depth + 1));
                            visited.insert(callee);
                        }
                    }
                }

                for (const auto &pair : depths) {
                    const std::string &func = pair.first;
                    unsigned depth = pair.second;
                    maxCallDepths[func] = std::max(maxCallDepths[func], depth);
                }

                LOG_DEBUG("calculateCallDepths", 
                    "Function {0}: processed {1} related functions, max depth {2}", 
                    startFunc, 
                    depths.size(),
                    maxCallDepths[startFunc]);
            }

            unsigned maxDepth = 0;
            std::string deepestFunc;
            for (const auto &pair : maxCallDepths) {
                if (pair.second > maxDepth) {
                    maxDepth = pair.second;
                    deepestFunc = pair.first;
                }
            }
            LOG_INFO("calculateCallDepths", 
                "Analysis complete - Total functions: {0}, Maximum depth: {1} (in function: {2})", 
                maxCallDepths.size(), 
                maxDepth,
                deepestFunc);
        }

        void generateProjectCallGraph() {
            errs() << "```mermaid\n";
            errs() << "graph TD\n";
            errs() << "    %% Project Call Graph\n";
            
            for (const auto &pair : callGraph) {
                const std::string &name = pair.first;
                const CallNode &node = pair.second;
                std::string nodeId = sanitizeNodeId(name);
                std::string depth = std::to_string(maxCallDepths[name]);
                std::string criticalPoints = std::to_string(node.criticalPoints.size());
                std::string style = node.isTarget ? ":::target" : "";
                errs() << "    " << nodeId << "[\"" << name 
                    << "\\nDepth: " << depth 
                    << "\\nCritical Points: " << criticalPoints
                    << "\"]" << style << "\n";
            }
            
            for (const auto &pair : callGraph) {
                const std::string &name = pair.first;
                const CallNode &node = pair.second;
                std::string callerId = sanitizeNodeId(name);
                for (const auto &callee : node.callees) {
                    std::string calleeId = sanitizeNodeId(callee);
                    errs() << "    " << callerId << " --> " << calleeId << "\n";
                }
            }
            errs() << "    classDef target fill:#f96,stroke:#333,stroke-width:4px\n";
            errs() << "```\n\n";
        }

        std::string sanitizeNodeId(const std::string &name) {
            std::string id = name;
            // 移除前导 %
            if (!id.empty() && id[0] == '%') {
                id = id.substr(1);
            }
            
            // 替换非法字符
            std::replace(id.begin(), id.end(), '.', '_');
            std::replace(id.begin(), id.end(), ' ', '_');
            std::replace(id.begin(), id.end(), '-', '_');
            std::replace(id.begin(), id.end(), '%', '_');
            
            // 确保ID以字母开头
            if (!id.empty() && std::isdigit(id[0])) {
                id = "block_" + id;
            }
            
            return id;
        }
    };
}

char CodeFusionPass::ID = 0;
static RegisterPass<CodeFusionPass> X("codefusion", "Code Fusion Pass");